/***************************************************************************
 main.cpp  -  description
 -------------------
 begin                : Fri Jan 17 17:59:14 UTC 2003
 copyright            : (C) 2003 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <iterator>
#include <memory>

#include <stdlib.h>
#include <string.h>
//#include "BioSequence.h"
#include "PairwiseAlignment.h"
#include "FarrarSWAligner.h"
#include "MonardoSWAligner.h"
#include "IOWrapper.h"

#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace std;
using namespace bio;

/** ======================================================== */
void usage(char* exec, po::options_description od) {
   if (!exec)
      return;
   if (strrchr(exec, '/'))
      exec = strrchr(exec, '/') + 1;
   cerr << "NAME\n\t" << exec << " - Aligns all seqeunces in file1 to all sequences in file2.\n\n";
   cerr << "SYNOPSYS\n\t" << exec << " infile1(-) infile2 [outfile] [options]\n";
   cerr << od;
   cerr << "\nAUTHOR\n\tOleksiy Kohany, kohany@girinst.org\n";
#ifdef COPYRIGHT

   cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}

/** ======================================================== */
void showVersion(const char* exec) {
   if (!exec)
      return;
   if (strrchr(exec, '/'))
      exec = strrchr(exec, '/') + 1;
   cerr << exec;
#ifdef VERSION

   cerr << "\tversion " << VERSION;
#endif

   cerr << endl;
}

/** ======================================================== */
int main(int argc, char *argv[]) {
   string infile1;
   string infile2;
   string outfile;

   auto_ptr<SWAligner> pAligner;
   int iGI = 0;
   int iGX = 0;
   string mask;

   po::options_description generic("GENERIC OPTIONS");
   po::options_description settings("SETTINGS");
   po::options_description alignment("ALIGNMENT OPTIONS");
   po::options_description backend("BACKEND OPTIONS");
   po::options_description hidden("HIDDEN OPTIONS");

   generic.add_options()
   ("version,v", "print version string.")
   ("help", "produce help message.")
   //("debug", "will show debugging messages on execution process.")
   ;

   settings.add_options()
   ("protein,p", "will assume that sequences are proteins and BLOSUM62 matrix will be used for alignment.")
   ("msk", po::value<string>(&mask), "pattern that should not be included in calculating similarity, if found in bottom sequence.")
   ;

   alignment.add_options()
#ifdef HAVE_SSE2
   ("sse2", "uses modified implementation of Michael Farrar's striped SWAT algorithm for SSE2 extension of x86 CPUs. This is the default alignment method. Default matrix - xrepmask/3, gi=-10, gx=-2. Limitations: Maximum score can not accede 65535 so keep alignment matrix's scores low. With the default nucleotide matrix you can align over 21KB.")
#endif
   ("mon", "use Monardo alignment algorithm. It has logarithmic gap penalty though much slower than sse2 version.")
   ("gi", po::value<int>(&iGI), "gap initialization penalty")
   ("gx", po::value<int>(&iGX), "gap extension penalty")
   ("mat", po::value<string>(), "file name for alignment scoring matrix")
   ;

   backend.add_options()
   //("nothreads","do not use multithreading. Useful if for some reason you need to reduce CPU utilization to just 1 processor. By default if -mon or -sse2 option is used the program will attempt to utilize all available CPUs for multiple alignment stage. Phill Green's SWAT implementation is not thread safe so it uses just one CPU thread.")
   ;
   hidden.add_options()
   ("input-file1", po::value<string>(&infile1), "top sequence file.")
   ("input-file2", po::value<string>(&infile2), "bottom sequence file.")
   ("output-file", po::value<string>(&outfile)->default_value("-"), "output file. If '-' - stdout will be taken")
   ;

   po::options_description cmdline_options;
   cmdline_options.add(generic).add(settings).add(alignment).add(backend).add(hidden);

   po::options_description visible_options;
   visible_options.add(generic).add(settings).add(alignment).add(backend);

   po::positional_options_description p;
   p.add("input-file1", 1).add("input-file2", 1).add("output-file", 1);

   po::variables_map vm;
   po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(p).run(), vm);
   po::notify(vm);

   if (vm.count("version")) {
      showVersion(argv[0]);
      exit(0);
   }
   if (vm.count("input-file1") == 0 || vm.count("input-file2") == 0) {
      usage(argv[0], visible_options);
      exit(0);
   }
   if (vm.count("debug")) {
      //DEBUG   = true;
   }
   if (vm.count("mon")) {
      pAligner.reset(new MonardoSWAligner());
   }
   if (vm.count("sse2")) {
      pAligner.reset(new FarrarSWAligner());
   }
   if(pAligner.get() == NULL) {
#ifdef HAVE_SSE2
      pAligner.reset(new FarrarSWAligner());
#else
      pAligner.reset(new MonardoSWAligner());
#endif
   }

   if (vm.count("protein")) {
      PairwiseAlignment::ScoreMatrix_t matrix(PairwiseAlignment::ScoreMatrix_t::PROTALNMATRIX, false);
      PairwiseAlignment::setScoreMatrix(matrix);
      pAligner->setAlignmentMatrix(matrix);
   }
   if (vm.count("mat")) {
      PairwiseAlignment::loadScoreMatrix(vm["mat"].as<string> ().c_str());
      pAligner->loadAlignmentMatrix(vm["mat"].as<string> ().c_str());
   }

   if (iGI != 0) {
      pAligner->setGapIniPenalty(iGI);
   }
   if (iGX != 0) {
      pAligner->setGapExtPenalty(iGX);
   }

   IstreamWrapper is1(infile1.c_str());
   IFstreamWrapper is2(infile2.c_str());
   OstreamWrapper os(outfile.c_str());
   //========================================================
   NamedSubSequence sTOP;
   PairwiseAlignment aln;
   if (vm.count("msk")) {
      aln.addIgnoreMask(vm["msk"].as<string> ());
   }
   while (!(is1 >> sTOP).fail()) {
      Fragment topPos;
      if(!sTOP.getFragments().empty()) {
         topPos = sTOP.getFragments().front();
      }
      //pAligner->align(sTOP, istream_iterator<BioSubSequence>(is2), istream_iterator<BioSubSequence>(), ostream_iterator<PairwiseAlignment>(os));
      while (!(is2 >> aln.getBot()).fail()) {
         Fragment botPos;
         if(!aln.getBot().getFragments().empty()) {
            botPos = aln.getBot().getFragments().front();
         }
         aln.getTop() = sTOP;
         if (pAligner->align(aln)) {
            if (topPos.size() + botPos.size() == 0) {
               os << aln;
            } else {
               aln.printAlnHead(os);
               os.stream() << " < Local\n";
               if (topPos.size() > 0) {
                  aln.getTop().reposition(topPos);
               }
               if (botPos.size() > 0) {
                  aln.getBot().reposition(botPos);
               }
               aln.printAlnHead(os);
               os.stream() << " < Global\n\n";
               aln.printAlnBody(os);
            }
         } else {
            cerr << "No alignment\n";
         }
      }
      is2.stream().clear();
      is2.stream().seekg(0);
   }
   //========================================================
}
