/*
 * Fragment.h
 *
 *  Created on: Feb 19, 2010
 *      Author: kohany
 */
#include <string>
#include <iostream>
#include <stdlib.h>

#ifdef HAVE_PTHREAD
#include "concurrent_set.hpp"
#else
#include <set>
#endif

using namespace std;

#ifndef FRAGMENT_H_
#define FRAGMENT_H_

namespace bio {

   const int MAXDIST = 30;

   /**
    * Stores coordinates of a fragment of a sequence with it's name.
    * Coordinates start from 0.
    */
   class Fragment {
      const string* pName;
      ssize_t beg, end;
      bool direct;

#ifdef HAVE_PTHREAD
      static concurrent_set<string> cset_;
#else
      static set<string> cset_;
#endif
   public:
      Fragment();
      Fragment(const string& name, size_t beg, size_t end);
      Fragment(const Fragment& other);
      virtual ~Fragment();
      Fragment& operator =(const Fragment& other);
      void assign(const string& name, size_t beg, size_t end);
      void assign(size_t beg, size_t end);
      void validate();
      const string& getName() const;
      void setName(const string& name) {
         pName = &(*cset_.insert(name).first);
      }
      const size_t getBeg() const {
         return beg;
      }
      const size_t getEnd() const {
         return end;
      }
      const size_t getRBeg() const {
         return direct ? beg : end;
      }
      const size_t getREnd() const {
         return direct ? end : beg;
      }
      void setBeg(ssize_t b) {
         beg = b;
      }
      void setEnd(ssize_t e) {
         end = e;
      }
      const bool isDirect() const {
         return direct;
      }
      void setDirect(bool direct) {
         this->direct = direct;
      }
      void flop() {
         direct = !direct;
      }
      const size_t size() const {
         return abs(end - beg);
      }
      const ssize_t getLength() const {
         return end - beg;
      }
      virtual void reset();
      virtual Fragment& operator +=(const Fragment& other);
      virtual Fragment& operator -=(const Fragment& other);
      virtual Fragment& operator +=(const size_t v);
      virtual Fragment& operator -=(const size_t v);
      virtual Fragment& operator *=(const double v);
      virtual Fragment& operator /=(const double v);
      virtual Fragment& reposition(const Fragment& parent);
      virtual pair<Fragment, Fragment> subsplit(const Fragment& other) const;
      virtual const Fragment operator -(const Fragment& other) const {
         Fragment res(*this);
         return (res -= other);
      }
      virtual const Fragment operator +(const Fragment& other) const {
         Fragment res(*this);
         return (res += other);
      }
      void resize(ssize_t left, ssize_t right);
      void expand(size_t left, size_t right);
      void shrink(size_t left, size_t right);
      virtual const bool equals(const Fragment& other) const;
      const bool operator ==(const Fragment& other) const {
         return equals(other);
      }
      virtual Fragment& read(const string& s);
      virtual Fragment& read(istream& is);
      virtual void write(string& s) const;
      virtual void write(ostream& os) const;
   };
   /**
    * Returns distance between two Fragments. If they overlap, returns
    * negative value equal to length of the overlap.
    */
   const ssize_t dist(const Fragment& f1, const Fragment& f2, bool compareNames = true);
   const ssize_t range(const Fragment& f1, const Fragment& f2, bool compareNames = true);
   const bool overlap(const Fragment& f1, const Fragment& f2, double threshold = 0, bool compareNames = true);
   const bool contain(const Fragment& f1, const Fragment& f2, bool compareNames = true);
   const bool areJoinable(const Fragment& f1, const Fragment& f2, int maxDist = MAXDIST, bool compareNames = true);
   const bool areOrdered(const Fragment& f1, const Fragment& f2, bool compareNames = true);
   istream& operator >>(istream& is, Fragment& f);
   ostream& operator <<(ostream& os, const Fragment& f);
}
#endif /* FRAGMENT_H_ */
