/*
 * PairwiseAlignment.h
 *
 *  Created on: Apr 19, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#ifndef PAIRWISEALIGNMENT_H_
#define PAIRWISEALIGNMENT_H_

#include <vector>
#include <iterator>
#include <assert.h>
#include "NamedSubSequence.h"
#include "ScoreMatrix.hpp"
using namespace std;
/*
 *
 */
namespace bio {
   struct PairwiseAlignmentStats {
      typedef size_t count_t;
      count_t iMMCnt; //Missmatch
      count_t iTRCnt; //Transition
      count_t iTVCnt; //Transversion
      count_t iMCnt; //Matches
      count_t iPCnt; //Positives
      count_t iGCnt; //Gap count
      PairwiseAlignmentStats() :
         iMMCnt(0), iTRCnt(0), iTVCnt(0), iMCnt(0), iPCnt(0), iGCnt(0) {
      }
      void reset() {
         iMMCnt = iTRCnt = iTVCnt = iMCnt = iPCnt = iGCnt = 0;
      }
      double getSim() {
         return (iMCnt + iMMCnt > 0) ? double(iMCnt) / (iMCnt + iMMCnt) : 0;
      }
      double getPos() {
         return (iMCnt + iMMCnt > 0) ? double(iPCnt) / (iMCnt + iMMCnt) : 0;
      }
      double getTrs() {
         return (double) iMMCnt / (1 + iTRCnt);
      }

   };



   class PairwiseAlignment {
   public:
      typedef ScoreMatrix<signed char> ScoreMatrix_t;
   private:
      typedef int score_t;
      NamedSubSequence top;
      NamedSubSequence bot;
      score_t score;
      static vector<string> ignoreMasks;
      static ScoreMatrix_t scoreMatrix;
      static string scoreMatrixName;
   public:
      PairwiseAlignment();
      PairwiseAlignment(const PairwiseAlignment& other);
      PairwiseAlignment(const NamedSubSequence& top, const NamedSubSequence& bot, score_t score = 0);
      virtual ~PairwiseAlignment();
      PairwiseAlignment& operator=(const PairwiseAlignment& other) {
         top = other.getTop();
         bot = other.getBot();
         return *this;
      }
      NamedSubSequence& getTop() {
         return top;
      }
      NamedSubSequence& getBot() {
         return bot;
      }
      const NamedSubSequence& getTop() const {
         return top;
      }
      const NamedSubSequence& getBot() const {
         return bot;
      }
      size_t size() {
         assert(top.size() == bot.size());
         return top.size();
      }
      score_t getScore() const {
         return score;
      }
      void setScore(score_t v) {
         score = v;
      }
      void addIgnoreMask(const string& mask) {
         ignoreMasks.push_back(mask);
      }
      void printAlnHead(ostream& os) const;
      void printAlnBody(ostream& os, size_t width = 70, const char* sBlahS = NULL, const char* sBlahQ = NULL) const;
      static void setScoreMatrix(const ScoreMatrix_t& matrix);
      static void loadScoreMatrix(const char* s);
      static const ScoreMatrix_t& getScoreMatrix() {
         return scoreMatrix;
      }
      PairwiseAlignmentStats getStat();
   };

   istream& operator>>(istream& is, PairwiseAlignment& aln);
   ostream& operator<<(ostream& os, const PairwiseAlignment& aln);
   ostream& operator<<(ostream& os, const PairwiseAlignmentStats& v);

   bool isTransition(char c1, char c2);
   char getLink(char c1, char c2, PairwiseAlignment::ScoreMatrix_t& matrix);
   PairwiseAlignment& sharpen(PairwiseAlignment& aln, size_t edgeWidth);
   PairwiseAlignment& trimLeft(PairwiseAlignment& aln, size_t trimLength);
   PairwiseAlignment& trimRight(PairwiseAlignment& aln, size_t trimLength);
   /**
    * ForwardIterator points to char
    * MasklistIterator points to string: mskBeg-mskEnd contains string masks which should be excluded from calculating stats.
    */
   template<typename ForwardIterator, typename MasklistIterator>
   PairwiseAlignmentStats& accumulateAlignmentStats(ForwardIterator topBeg, ForwardIterator topEnd, ForwardIterator botBeg,
                                                    ForwardIterator botEnd, MasklistIterator mskBeg, MasklistIterator mskEnd,
                                                    const typename PairwiseAlignment::ScoreMatrix_t& matrix, PairwiseAlignmentStats& stat) {
      unsigned int iCGapLen = 0;
      unsigned int iSGapLen = 0;
      ForwardIterator topPtr, botPtr;
      for (topPtr = topBeg, botPtr = botBeg; topPtr != topEnd && botPtr != botEnd; topPtr++, botPtr++) {
         if (*botPtr == matrix.GAPCHAR && *topPtr == matrix.GAPCHAR) {
            continue;
         }
         if (*botPtr == matrix.LPADDING || *topPtr == matrix.LPADDING || *botPtr == matrix.RPADDING || *topPtr == matrix.RPADDING) {
            continue;
         }
         bool skip = false;
         for (MasklistIterator mskPtr = mskBeg; mskPtr != mskEnd; mskPtr++) {
            typename iterator_traits<MasklistIterator>::value_type::const_iterator sptr2;
            ForwardIterator sptr1;
            for (sptr1 = botPtr, sptr2 = (*mskPtr).begin(); sptr1 != botEnd && sptr2 != (*mskPtr).end(); sptr1 = firstBase(++sptr1, botEnd), sptr2++) {
               if (toupper(*sptr1) != toupper(*sptr2)) {
                  break;
               }
            }
            if (sptr2 == (*mskPtr).end()) {
               skip = true;
               while(botPtr != sptr1) {
                  ++topPtr;
                  ++botPtr;
               }
               //topPtr += distance(botPtr, sptr1) - 1;
               //botPtr += distance(botPtr, sptr1) - 1;
               break;
            }
         }
         if (skip) {
            cerr << "Skipping ingnoreMask match"<<endl;
            continue;
         }

         if (*botPtr != *topPtr) {
            if (*botPtr == matrix.GAPCHAR) {
               iCGapLen++;
               continue;
            }
            if (*topPtr == matrix.GAPCHAR) {
               iSGapLen++;
               continue;
            }
         } else {
            if (*botPtr == matrix.GAPCHAR)
               continue;
         }
         if (iCGapLen > 0) {
            stat.iMMCnt++;
            stat.iGCnt++;
            iCGapLen = 0;
         }
         if (iSGapLen > 0) {
            stat.iMMCnt++;
            stat.iGCnt++;
            iSGapLen = 0;
         }
         if (matrix.getScore(*botPtr, *topPtr) > 0) {
            stat.iPCnt++;
         }
         if (matrix.getChar2Index()[*topPtr] == matrix.getChar2Index()[*botPtr] && matrix.getScore(*botPtr, *topPtr) > 0) {
            stat.iMCnt++;
         } else {
            stat.iMMCnt++;
            if (matrix.getType() == NA) {
               if (isTransition(*botPtr, *topPtr)) {
                  stat.iTRCnt++;
               } else {
                  stat.iTVCnt++;
               }
            }
         }
      }
      return stat;
   }

   template<typename InputIterator, typename OutputIterator>
   double accumulateScore(InputIterator beg1, InputIterator end1, InputIterator beg2, InputIterator end2, OutputIterator obeg,
                          OutputIterator oend, const typename PairwiseAlignment::ScoreMatrix_t& matrix, bool ignoreGaps = false) {
      double sumScore = 0;
      bool isInGap = false;
      OutputIterator optr = obeg;
      for (InputIterator iptr1 = beg1, iptr2 = beg2; iptr1 != end1 && iptr2 != end2; ++iptr1, ++iptr2) {
         double rowScore = 0;
         double colScore = 0;
         if (*iptr1 == matrix.GAPCHAR || *iptr2 == matrix.GAPCHAR) {
            if (!ignoreGaps) {
               colScore = matrix.getScore(*iptr1, *iptr2);
               if (*iptr1 != *iptr2) {
                  if (isInGap) {
                     rowScore = matrix.getGapExtPenalty();
                  } else {
                     rowScore = matrix.getGapIniPenalty();
                     isInGap = true;
                  }
               } else {
                  isInGap = false;
               }
            }
         } else {
            isInGap = false;
            rowScore = colScore = matrix.getScore(*iptr1, *iptr2);
         }
         sumScore += rowScore;
         if (optr != oend) {
            *optr++ += colScore;
         }
      }
      return sumScore;
   }

}

#endif /* PAIRWISEALIGNMENT_H_ */
