/*
 * SWAligner.h
 *
 *  Created on: Apr 6, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#ifndef SWALIGNER_H_
#define SWALIGNER_H_

#include <assert.h>
#include <stdlib.h>
#include "PairwiseAlignment.h"

#include <memory>
#include <algorithm>
using namespace std;
/*
 *
 */
namespace bio {
   class SwData {
      void *pData;
   protected:
      void* alloc(size_t size) {
         dealloc();
         pData = calloc(size, 1);
         assert(pData);
         return pData;
      }
      SwData() :
         pData(NULL) {
      }
   public:
      virtual ~SwData() {
         dealloc();
      }
      void dealloc() {
         if (pData != NULL) {
            free(pData);
         }
      }
   };

   class SWAligner {
      static const char lightWeightMatrix[];
   public:
      typedef signed char score_t;
      typedef ScoreMatrix<score_t> ScoreMatrix_t;
   protected:
      ScoreMatrix_t scoreMatrix;
      typedef vector<unsigned char> vindex_t;
      typedef vector<short> vtrace_t;
      void printAlignmentMatrix(const short* scoreMat, int height, int width, const char* rowHead, const char* colHead);
      void printTraceMatrix(const short* scoreMat, int height, int width, const char* rowHead, const char* colHead);
      virtual int traceBack(const vindex_t& topSeq, const vindex_t& botSeq, const vtrace_t& pathMat, int iMaxX, int iMaxY,
                            SubSequence& top, SubSequence& bot) const;
      virtual auto_ptr<SwData> getProfile(const vindex_t& s) const = 0;
      virtual double
            alignmentImplementation(const vindex_t& topSeq, const vindex_t& botSeq, unsigned short gapOpen, unsigned short gapExtend,
                                    SwData *topProfile, SubSequence& top, SubSequence& bot) const = 0;
   public:
      SWAligner();
      virtual ~SWAligner();
      vindex_t& string2vector(const string& s, vindex_t& v) const;
      string& vector2string(const vindex_t& v, string& s) const;
      void setGapIniPenalty(score_t penalty) {
         scoreMatrix.setGapIniPenalty(penalty);
      }
      void setGapExtPenalty(score_t penalty) {
         scoreMatrix.setGapExtPenalty(penalty);
      }
      void loadAlignmentMatrix(const char *file) {
         scoreMatrix.parse(file);
      }
      void setAlignmentMatrix(const ScoreMatrix_t& matrix) {
         scoreMatrix = matrix;
      }
      double align(SubSequence& top, SubSequence& bot) const;
      double align(PairwiseAlignment& aln) const {
         aln.setScore(align(aln.getTop(), aln.getBot()));
         return aln.getScore();
      }
      template<class InputIterator, class OutputIterator>
      double align(NamedSubSequence& top, InputIterator botBeg, InputIterator botEnd, OutputIterator optr) const;
   };

   template<class InputIterator, class OutputIterator>
   double SWAligner::align(NamedSubSequence& top, InputIterator botBeg, InputIterator botEnd, OutputIterator optr) const {
      vindex_t vtop;
      vindex_t vbot;
      PairwiseAlignment bioAln;
      bioAln.getTop().setName(top.getName());
      transform(top.begin(), top.end(), back_insert_iterator<vindex_t>(vtop), std::bind1st(std::mem_fun(&ScoreMatrix_t::c2i), &scoreMatrix));
      //string2vector(top, vtop);
      auto_ptr<SwData> pTopProfile(getProfile(vtop));
      for (InputIterator sptr = botBeg; sptr != botEnd; sptr++) {
         string2vector(*sptr, vbot);
         double score = alignmentImplementation(vtop, vbot, scoreMatrix.getGapIniPenalty(), scoreMatrix.getGapExtPenalty(),
                                                pTopProfile.get(), bioAln.getTop(), bioAln.getBot());
         if (score > 0) {
            bioAln.getBot().setName((*sptr).getName());
            *optr = bioAln;
            ++optr;
         }
      }
   }
}
#endif /* SWALIGNER_H_ */
